@@ -102,6 +102,7 @@ class AsyncClient(client.Client):
                                    engineio_path=socketio_path)
         except engineio.exceptions.ConnectionError as exc:
             six.raise_from(exceptions.ConnectionError(exc.args[0]), None)
+        self.connected = True
 
     async def wait(self):
         """Wait until the connection with the server ends.
@@ -141,6 +142,9 @@ class AsyncClient(client.Client):
         Note: this method is a coroutine.
         """
         namespace = namespace or '/'
+        if namespace != '/' and namespace not in self.namespaces:
+            raise exceptions.BadNamespaceError(
+                namespace + ' is not a connected namespace.')
         self.logger.info('Emitting event "%s" [%s]', event, namespace)
         if callback is not None:
             id = self._generate_ack_id(namespace, callback)
@@ -185,18 +189,54 @@ class AsyncClient(client.Client):
         await self.emit('message', data=data, namespace=namespace,
                         callback=callback)
 
+    async def call(self, event, data=None, namespace=None, timeout=60):
+        """Emit a custom event to a client and wait for the response.
+
+        :param event: The event name. It can be any string. The event names
+                      ``'connect'``, ``'message'`` and ``'disconnect'`` are
+                      reserved and should not be used.
+        :param data: The data to send to the client or clients. Data can be of
+                     type ``str``, ``bytes``, ``list`` or ``dict``. If a
+                     ``list`` or ``dict``, the data will be serialized as JSON.
+        :param namespace: The Socket.IO namespace for the event. If this
+                          argument is omitted the event is emitted to the
+                          default namespace.
+        :param timeout: The waiting timeout. If the timeout is reached before
+                        the client acknowledges the event, then a
+                        ``TimeoutError`` exception is raised.
+
+        Note: this method is a coroutine.
+        """
+        callback_event = self.eio.create_event()
+        callback_args = []
+
+        def event_callback(*args):
+            callback_args.append(args)
+            callback_event.set()
+
+        await self.emit(event, data=data, namespace=namespace,
+                        callback=event_callback)
+        try:
+            await asyncio.wait_for(callback_event.wait(), timeout)
+        except asyncio.TimeoutError:
+            six.raise_from(exceptions.TimeoutError(), None)
+        return callback_args[0] if len(callback_args[0]) > 1 \
+            else callback_args[0][0] if len(callback_args[0]) == 1 \
+            else None
+
     async def disconnect(self):
         """Disconnect from the server.
 
         Note: this method is a coroutine.
         """
+        # here we just request the disconnection
+        # later in _handle_eio_disconnect we invoke the disconnect handler
         for n in self.namespaces:
-            await self._trigger_event('disconnect', namespace=n)
             await self._send_packet(packet.Packet(packet.DISCONNECT,
                                     namespace=n))
-        await self._trigger_event('disconnect', namespace='/')
         await self._send_packet(packet.Packet(
             packet.DISCONNECT, namespace='/'))
+        self.connected = False
         await self.eio.disconnect(abort=True)
 
     def start_background_task(self, target, *args, **kwargs):
@@ -251,10 +291,18 @@ class AsyncClient(client.Client):
             self.namespaces.append(namespace)
 
     async def _handle_disconnect(self, namespace):
+        if not self.connected:
+            return
         namespace = namespace or '/'
+        if namespace == '/':
+            for n in self.namespaces:
+                await self._trigger_event('disconnect', namespace=n)
+            self.namespaces = []
         await self._trigger_event('disconnect', namespace=namespace)
         if namespace in self.namespaces:
             self.namespaces.remove(namespace)
+        if namespace == '/':
+            self.connected = False
 
     async def _handle_event(self, namespace, id, data):
         namespace = namespace or '/'
@@ -300,6 +348,9 @@ class AsyncClient(client.Client):
             namespace))
         if namespace in self.namespaces:
             self.namespaces.remove(namespace)
+        if namespace == '/':
+            self.namespaces = []
+            self.connected = False
 
     async def _trigger_event(self, event, namespace, *args):
         """Invoke an application event handler."""
@@ -320,6 +371,8 @@ class AsyncClient(client.Client):
                 event, *args)
 
     async def _handle_reconnect(self):
+        self._reconnect_abort.clear()
+        client.reconnecting_clients.append(self)
         attempt_count = 0
         current_delay = self.reconnection_delay
         while True:
@@ -331,12 +384,18 @@ class AsyncClient(client.Client):
             self.logger.info(
                 'Connection failed, new attempt in {:.02f} seconds'.format(
                     delay))
-            await self.sleep(delay)
+            try:
+                await asyncio.wait_for(self._reconnect_abort.wait(), delay)
+                self.logger.info('Reconnect task aborted')
+                break
+            except (asyncio.TimeoutError, asyncio.CancelledError):
+                pass
             attempt_count += 1
             try:
                 await self.connect(self.connection_url,
                                    headers=self.connection_headers,
                                    transports=self.connection_transports,
+                                   namespaces=self.connection_namespaces,
                                    socketio_path=self.socketio_path)
             except (exceptions.ConnectionError, ValueError):
                 pass
@@ -349,10 +408,12 @@ class AsyncClient(client.Client):
                 self.logger.info(
                     'Maximum reconnection attempts reached, giving up')
                 break
+        client.reconnecting_clients.remove(self)
 
-    def _handle_eio_connect(self):  # pragma: no cover
+    def _handle_eio_connect(self):
         """Handle the Engine.IO connection event."""
         self.logger.info('Engine.IO connection established')
+        self.sid = self.eio.sid
 
     async def _handle_eio_message(self, data):
         """Dispatch Engine.IO messages."""
@@ -385,11 +446,16 @@ class AsyncClient(client.Client):
     async def _handle_eio_disconnect(self):
         """Handle the Engine.IO disconnection event."""
         self.logger.info('Engine.IO connection dropped')
-        for n in self.namespaces:
-            await self._trigger_event('disconnect', namespace=n)
-        await self._trigger_event('disconnect', namespace='/')
+        self._reconnect_abort.set()
+        if self.connected:
+            for n in self.namespaces:
+                await self._trigger_event('disconnect', namespace=n)
+            await self._trigger_event('disconnect', namespace='/')
+            self.namespaces = []
+            self.connected = False
         self.callbacks = {}
         self._binary_packet = None
+        self.sid = None
         if self.eio.state == 'connected' and self.reconnection:
             self._reconnect_task = self.start_background_task(
                 self._handle_reconnect)
